@@ -5,10 +5,16 @@ Google Cloud Storage pythonic interface
 from __future__ import print_function
 
 import array
+from base64 import b64encode
+import google.auth as gauth
+from google.auth.transport.requests import AuthorizedSession
+from google.oauth2.credentials import Credentials
+from google_auth_oauthlib.flow import InstalledAppFlow
+from google.oauth2 import service_account
+from hashlib import md5
 import io
 import json
 import logging
-import oauth2client.client   # version 1.5.2
 import os
 import pickle
 import re
@@ -16,17 +22,26 @@ import requests
 import sys
 import time
 import warnings
-import webbrowser
 
+from requests.exceptions import RequestException
+from .utils import HtmlError
+from .utils import is_retriable
 from .utils import read_block
+
 PY2 = sys.version_info.major == 2
 
 logger = logging.getLogger(__name__)
 
 # client created 23-Sept-2017
-not_secret = {"client_id": "586241054156-7a3vrghs70ffkkfkmnnatjnbjg03cq9a."
-                           "apps.googleusercontent.com",
-              "client_secret": "RsSQZJKYE00oRv2ibYlj28Sx"}
+not_secret = {"client_id": "586241054156-0asut23a7m10790r2ik24309flribp7j"
+                           ".apps.googleusercontent.com",
+              "client_secret": "w6VkI99jS6e9mECscNztXvQv"}
+client_config = {'installed': {
+    'client_id': not_secret['client_id'],
+    'client_secret': not_secret['client_secret'],
+    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
+    "token_uri": "https://accounts.google.com/o/oauth2/token"
+}}
 tfile = os.path.join(os.path.expanduser("~"), '.gcs_tokens')
 ACLs = {"authenticatedread", "bucketownerfullcontrol", "bucketownerread",
         "private", "projectprivate", "publicread"}
@@ -77,6 +92,8 @@ def split_path(path):
         path = path[6:]
     if path.startswith('gs://'):
         path = path[5:]
+    if path.startswith('/'):
+        path = path[1:]
     if '/' not in path:
         return path, ""
     else:
@@ -94,8 +111,10 @@ def validate_response(r, path):
     """
     if not r.ok:
         m = str(r.content)
+        error = None
         try:
-            msg = r.json()['error']['message']
+            error = r.json()['error']
+            msg = error['message']
         except:
             msg = str(r.content)
 
@@ -107,6 +126,8 @@ def validate_response(r, path):
             raise IOError("Forbidden: %s\n%s" % (path, msg))
         elif "invalid" in m:
             raise ValueError("Bad Request: %s\n%s" % (path, msg))
+        elif error:
+            raise HtmlError(error)
         else:
             raise RuntimeError(m)
 
@@ -115,54 +136,72 @@ class GCSFileSystem(object):
     """
     Connect to Google Cloud Storage.
 
-    Two modes of authentication are supported:
-    - if ``token=None``, you will be given a "device code", which you must
-      enter into a browser where you are logged in with your Google identity.
+    The following modes of authentication are supported:
+    - ``token=None``, GCSFS will attempt to guess your credentials in the
+      following order: gcloud CLI default, gcsfs cached token, google compute
+      metadata service, anonymous.
+    - ``token='google_default'``, your default gcloud credentials will be used,
+      which are typically established by doing ``gcloud login`` in a terminal.
+    - ``token=='cache'``, credentials from previously successful gcsfs
+      authentication will be used (use this after "browser" auth succeeded)
+    - ``token='anon'``, no authentication is preformed, and you can only
+      access data which is accessible to allUsers (in this case, the project and
+      access level parameters are meaningless)
+    - ``token='browser'``, you get an access code with which you can
+      authenticate via a specially provided URL
     - if ``token='cloud'``, we assume we are running within google compute
       or google container engine, and query the internal metadata directly for
       a token.
     - you may supply a token generated by the
       [gcloud](https://cloud.google.com/sdk/docs/)
-      utility; this is either a python dictionary, or the name of a file
-      containing the JSON returned by logging in with the gcloud CLI tool. On
-      a posix system this may be at
-      ``~/.config/gcloud/application_default_credentials.json``
-
-    We maintain a cache of refresh tokens in the file ~/.gcs_tokens, so for any
-    pair of (project, access), you will not need to log in once your credentials
-    are verified.
+      utility; this is either a python dictionary, the name of a file
+      containing the JSON returned by logging in with the gcloud CLI tool,
+      or a Credentials object. gcloud typically stores its tokens in locations
+      such as
+      ``~/.config/gcloud/application_default_credentials.json``,
+      `` ~/.config/gcloud/credentials``, or
+      ``~\AppData\Roaming\gcloud\credentials``, etc.
 
     Parameters
     ----------
     project : string
-        GCS users may only access to contents of one project in a single
-        instance of GCSFileSystem. This is required in order
-        to list all the buckets you have access to within a project.
+        project_id to work under. Note that this is not the same as, but ofter
+        very similar to, the project name.
+        This is required in order
+        to list all the buckets you have access to within a project and to
+        create/delete buckets, or update their access policies.
+        If ``token='google_default'``, the value is overriden by the default,
+        if ``token='anon'``, the value is ignored.
     access : one of {'read_only', 'read_write', 'full_control'}
         Full control implies read/write as well as modifying metadata,
         e.g., access control.
     token: None, dict or string
         (see description of authentication methods, above)
+    consistency: 'none', 'size', 'md5'
+        Check method when writing files. Can be overridden in open().
     """
     scopes = {'read_only', 'read_write', 'full_control'}
-    retries = 10
+    retries = 4  # number of retries on http failure
     base = "https://www.googleapis.com/storage/v1/"
     _singleton = [None]
     default_block_size = 5 * 2**20
 
     def __init__(self, project=DEFAULT_PROJECT, access='full_control',
-                 token=None, block_size=None):
+                 token=None, block_size=None, consistency='none'):
         if access not in self.scopes:
             raise ValueError('access must be one of {}', self.scopes)
         if project is None:
             warnings.warn('GCS project not set - cannot list or create buckets')
-        self.input_token = token
         if block_size is not None:
             self.default_block_size = block_size
         self.project = project
         self.access = access
+        self.scope = "https://www.googleapis.com/auth/devstorage." + access
+        self.consistency = consistency
         self.dirs = {}
-        self.connect()
+        self.token = token
+        self.session = None
+        self.connect(method=token)
         self._singleton[0] = self
 
     @classmethod
@@ -176,108 +215,117 @@ class GCSFileSystem(object):
         else:
             return cls._singleton[0]
 
-    @staticmethod
-    def _parse_gtoken(gt):
-        if isinstance(gt, str):
-            t = json.load(open(gt))
-        else:
-            t = gt.copy()
-        typ = t.pop('type')
-        if typ != "authorized_user":
-            raise ValueError("Only 'authorized_user' tokens accepted, got: %s"
-                             % typ)
-        t['grant_type'] = 'refresh_token'
-        t['timestamp'] = time.time()
-        t['expires_in'] = 0
-        return t
-
     @staticmethod
     def load_tokens():
         try:
             with open(tfile, 'rb') as f:
                 tokens = pickle.load(f)
-        except Exception:
+            # backwards compatability
+            tokens = {k: (GCSFileSystem._dict_to_credentials(v)
+                          if isinstance(v, dict) else v)
+                      for k, v in tokens.items()}
+        except IOError:
             tokens = {}
         GCSFileSystem.tokens = tokens
 
-    def connect(self, refresh=False):
+    def _connect_google_default(self):
+        credentials, project = gauth.default()
+        self.project = project
+        self.session = AuthorizedSession(credentials)
+
+    def _connect_cloud(self):
+        credentials = gauth.compute_engine.Credentials()
+        self.session = AuthorizedSession(credentials)
+
+    def _connect_cache(self):
+        project, access = self.project, self.access
+        if (project, access) in self.tokens:
+            credentials = self.tokens[(project, access)]
+            self.session = AuthorizedSession(credentials)
+
+    @staticmethod
+    def _dict_to_credentials(token):
+        """
+        Convert old dict-style token.
+
+        Does not preserve access token itself, assumes refresh required.
+        """
+        return Credentials(
+            None, refresh_token=token['refresh_token'],
+            client_secret=token['client_secret'],
+            client_id=token['client_id'],
+            token_uri='https://www.googleapis.com/oauth2/v4/token'
+        )
+
+    def _connect_token(self, token):
+        """
+        Connect using a concrete token
+
+        Parameters
+        ----------
+        token: str, dict or Credentials
+            If a str, try to load as a Service file, or next as a JSON; if
+            dict, try to interpret as credentials; if Credentials, use directly.
+        """
+        if isinstance(token, str):
+            if not os.path.exists(token):
+                raise FileNotFoundError(token)
+            try:
+                # is this a "service" token?
+                self._connect_service(token)
+                return
+            except:
+                # some other kind of token file
+                # will raise exception if is not json
+                token = json.load(open(token))
+        if isinstance(token, dict):
+            credentials = GCSFileSystem._dict_to_credentials(token)
+        elif isinstance(token, Credentials):
+            credentials = token
+        else:
+            raise ValueError('Token format no understood')
+        self.session = AuthorizedSession(credentials)
+
+    def _connect_service(self, fn):
+        # raises exception if file does not match expectation
+        credentials = service_account.Credentials.from_service_account_file(fn)
+        self.session = AuthorizedSession(credentials)
+
+    def _connect_anon(self):
+        self.session = requests.Session()
+
+    def _connect_browser(self):
+        flow = InstalledAppFlow.from_client_config(client_config, [self.scope])
+        credentials = flow.run_console()
+        self.tokens[(self.project, self.access)] = credentials
+        self._save_tokens()
+        self.session = AuthorizedSession(credentials)
+
+    def connect(self, method=None):
         """
         Establish session token. A new token will be requested if the current
         one is within 100s of expiry.
 
         Parameters
         ----------
-        refresh: bool (False)
-            Force refresh, even if the token is expired.
+        method: str (google_default|cache|cloud|token|anon|browser) or None
+            Type of authorisation to implement - calls `_connect_*` methods.
+            If None, will try sequence of methods.
         """
-        token = self.input_token
-        project, access = self.project, self.access
-        if token is not None and str(token) != 'cloud':
-            if 'type' in token or isinstance(token, str):
-                token = self._parse_gtoken(token)
-            self.tokens[(project, access)] = token
-        if (project, access) in self.tokens:
-            # cached credentials
-            data = self.tokens[(project, access)]
-        elif token == 'cloud':
-            data = {'timestamp': time.time() - 3600, 'expires_in': 1,
-                    'type': 'cloud'}
+        if method not in ['google_default', 'cache', 'cloud', 'token', 'anon',
+                          'browser', None]:
+            self._connect_token(method)
+        elif method is None:
+            for meth in ['google_default', 'cache', 'cloud', 'anon']:
+                try:
+                    self.connect(method=meth)
+                except:
+                    logging.debug('Connection with method "%s" failed' % meth)
+                if self.session:
+                    break
         else:
-            # no credentials - try to ask google in the browser
-            scope = "https://www.googleapis.com/auth/devstorage." + access
-            path = 'https://accounts.google.com/o/oauth2/device/code'
-            r = requests.post(path,
-                              params={'client_id': not_secret['client_id'],
-                                      'scope': scope})
-            validate_response(r, path)
-            data = json.loads(r.content.decode())
-            print('Enter the following code when prompted in the browser:')
-            print(data['user_code'])
-            webbrowser.open(data['verification_url'])
-            for i in range(self.retries):
-                time.sleep(2)
-                r = requests.post(
-                        "https://www.googleapis.com/oauth2/v4/token",
-                        params={'client_id': not_secret['client_id'],
-                                'client_secret': not_secret['client_secret'],
-                                'code': data['device_code'],
-                                'grant_type':
-                                    "http://oauth.net/grant_type/device/1.0"})
-                data2 = json.loads(r.content.decode())
-                if 'error' in data2:
-                    if i == self.retries - 1:
-                        raise RuntimeError("Waited too long for browser"
-                                           "authentication.")
-                    continue
-                data = data2
-                break
-            data['timestamp'] = time.time()
-            data.update(not_secret)
-        if refresh or time.time() - data['timestamp'] > data['expires_in'] - 100:
-            # token has expired, or is about to - call refresh
-            if data.get('type', None) == 'cloud':
-                path = ('http://metadata.google.internal/computeMetadata/v1/'
-                        'instance/service-accounts/default/token')
-                r = requests.get(path, headers={'Metadata-Flavor': 'Google'})
-                validate_response(r, path)
-                data = r.json()
-                data['timestamp'] = time.time()
-                data['type'] = 'cloud'
-            else:
-                path = "https://www.googleapis.com/oauth2/v4/token"
-                r = requests.post(
-                        path,
-                        params={'client_id': data['client_id'],
-                                'client_secret': data['client_secret'],
-                                'refresh_token': data['refresh_token'],
-                                'grant_type': "refresh_token"})
-                validate_response(r, path)
-                data['timestamp'] = time.time()
-                data['access_token'] = r.json()['access_token']
-
-        self.tokens[(project, access)] = data
-        self.header = {'Authorization': 'Bearer ' + data['access_token']}
-        self._save_tokens()
+            self.__getattribute__('_connect_' + method)()
+            self.method = method
 
     @staticmethod
     def _save_tokens():
@@ -293,22 +341,35 @@ class GCSFileSystem(object):
             if v is None:
                 del kwargs[k]
         json = kwargs.pop('json', None)
-        meth = getattr(requests, method)
+        meth = getattr(self.session, method)
         if args:
             path = path.format(*[quote_plus(p) for p in args])
-        r = meth(self.base + path, headers=self.header, params=kwargs,
-                 json=json)
+        for retry in range(self.retries):
+            try:
+                time.sleep(2**retry - 1)
+                r = meth(self.base + path, params=kwargs, json=json)
+                validate_response(r, path)
+                break
+            except (HtmlError, RequestException) as e:
+                if retry == self.retries - 1:
+                    raise e
+                if is_retriable(e):
+                    # retry
+                    continue
+                raise e
         try:
             out = r.json()
         except ValueError:
             out = r.content
-        validate_response(r, path)
         return out
 
     def _list_buckets(self):
         if '' not in self.dirs:
-            out = self._call('get', 'b/', project=self.project)
-            dirs = out.get('items', [])
+            try:
+                out = self._call('get', 'b/', project=self.project)
+                dirs = out.get('items', [])
+            except (FileNotFoundError, IOError, ValueError):
+                dirs = []
             self.dirs[''] = dirs
         return self.dirs['']
 
@@ -318,7 +379,9 @@ class GCSFileSystem(object):
             dirs = out.get('items', [])
             next_page_token = out.get('nextPageToken', None)
             while next_page_token is not None:
-                out = self._call('get', 'b/{}/o/', bucket, maxResults=max_results, pageToken=next_page_token)
+                out = self._call('get', 'b/{}/o/', bucket,
+                                 maxResults=max_results,
+                                 pageToken=next_page_token)
                 dirs.extend(out.get('items', []))
                 next_page_token = out.get('nextPageToken', None)
             for f in dirs:
@@ -354,7 +417,7 @@ class GCSFileSystem(object):
             for v in self.dirs[''][:]:
                 if v['name'] == bucket:
                     self.dirs[''].remove(v)
-        self.invalidate_cache(bucket)
+        self.dirs.pop(bucket, None)
 
     def invalidate_cache(self, bucket=None):
         """
@@ -443,8 +506,8 @@ class GCSFileSystem(object):
             root = ''
         allfiles = self.walk(root)
         pattern = re.compile("^" + path.replace('//', '/')
-                             .rstrip('/')
-                             .replace('*', '[^/]*')
+                             .rstrip('/').replace('**', '.+')
+                             .replace('*', '[^/]+')
                              .replace('?', '.') + "$")
         out = [f for f in allfiles if re.match(pattern,
                f.replace('//', '/').rstrip('/'))]
@@ -456,18 +519,54 @@ class GCSFileSystem(object):
             if key:
                 return bool(self.info(path))
             else:
-                return bucket in self.ls('')
+                if bucket in self.ls(''):
+                    return True
+                else:
+                    try:
+                        self._list_bucket(bucket)
+                        return True
+                    except (FileNotFoundError, IOError, ValueError):
+                        # bucket listing failed as it doesn't exist or we can't
+                        # see it
+                        return False
         except FileNotFoundError:
             return False
 
     def info(self, path):
-        path = '/'.join(split_path(path))
-        files = self.ls(path, True)
-        out = [f for f in files if f['name'] == path]
+        bucket, key = split_path(path)
+        if not key:
+            files = self.ls('', True)
+            f = [f for f in files if f['name'] == bucket]
+            if f:
+                return f[0]
+            if self.ls(bucket):
+                return {'bucket': bucket, 'kind': 'storage#object',
+                        'size': 0, 'storageClass': 'DIRECTORY',
+                        'name': bucket+'/'}
+            raise FileNotFoundError
+        if bucket not in self.dirs:
+            try:
+                d = self._call('get', 'b/{}/o/{}', bucket, key)
+                d['name'] = '%s/%s' % (bucket, d['name'])
+                d['size'] = int(d.get('size'), 0)
+                return d
+            except FileNotFoundError:
+                pass
+        path1 = '/'.join(split_path(path))
+        out = []
+        try:
+            files = self.ls(path1, True)
+            out = [f for f in files if f['name'] == path1]
+        except FileNotFoundError:
+            pass
+        if not out:
+            # no such file, but try for such a directory
+            parent = path.rstrip('/').rsplit('/', 1)[0]
+            files = self.ls(parent, True)
+            out = [f for f in files if f['name'] == path.rstrip('/') + '/']
         if out:
             return out[0]
-        else:
-            raise FileNotFoundError(path)
+        raise FileNotFoundError(path)
 
     def url(self, path):
         return self.info(path)['mediaLink']
@@ -475,7 +574,7 @@ class GCSFileSystem(object):
     def cat(self, path):
         """ Simple one-shot get of file data """
         details = self.info(path)
-        return _fetch_range(self.header, details)
+        return _fetch_range(details, self.session)
 
     def get(self, rpath, lpath, blocksize=5 * 2 ** 20):
         with self.open(rpath, 'rb', block_size=blocksize) as f1:
@@ -526,15 +625,36 @@ class GCSFileSystem(object):
         self.copy(path1, path2, acl)
         self.rm(path1)
 
-    def rm(self, path):
-        bucket, path = split_path(path)
-        self._call('delete', "b/{}/o/{}", bucket, path)
-        self.invalidate_cache(bucket)
+    def rm(self, path, recursive=False):
+        """Delete keys. If recursive, also delete all keys
+        given by walk(path)"""
+        if recursive:
+            for p in self.walk(path):
+                self.rm(p)
+        else:
+            bucket, path = split_path(path)
+            self._call('delete', "b/{}/o/{}", bucket, path)
+            self.invalidate_cache(bucket)
 
-    def open(self, path, mode='rb', block_size=None, acl=None):
+    def open(self, path, mode='rb', block_size=None, acl=None,
+             consistency=None, metadata=None):
+        """
+        See ``GCSFile``.
+
+        consistency: None or str
+            If None, use default for this instance
+        """
         if block_size is None:
             block_size = self.default_block_size
-        return GCSFile(self, path, mode, block_size)
+        const = consistency or self.consistency
+        if 'b' in mode:
+            return GCSFile(self, path, mode, block_size, consistency=const,
+                           metadata=metadata)
+        else:
+            mode = mode.replace('t', '') + 'b'
+            return io.TextIOWrapper(
+                GCSFile(self, path, mode, block_size, consistency=const,
+                        metadata=metadata))
 
     def touch(self, path):
         with self.open(path, 'wb'):
@@ -588,13 +708,14 @@ class GCSFileSystem(object):
 
     def __getstate__(self):
         d = self.__dict__.copy()
-        del d['header']
+        del d['dirs']
         logger.debug("Serialize with state: %s", d)
         return d
 
     def __setstate__(self, state):
         self.__dict__.update(state)
-        self.connect()
+        self.dirs = {}
+        self.connect(self.token)
 
 
 GCSFileSystem.load_tokens()
@@ -603,11 +724,38 @@ GCSFileSystem.load_tokens()
 class GCSFile:
 
     def __init__(self, gcsfs, path, mode='rb', block_size=5 * 2 ** 20,
-                 acl=None):
+                 acl=None, consistency='md5', metadata=None):
+        """
+        Open a file.
+
+        Parameters
+        ----------
+        gcsfs: instance of GCSFileSystem
+        path: str
+            location in GCS, like 'bucket/path/to/file'
+        mode: str
+            Normal file modes. Currently only 'wb' amd 'rb'.
+        block_size: int
+            Buffer size for reading or writing
+        acl: str
+            ACL to apply, if any, one of ``ACLs``. New files are normally
+            "bucketownerfullcontrol", but a default can be configured per
+            bucket.
+        consistency: str, 'none', 'size', 'md5'
+            Check for success in writing, applied at file close.
+            'size' ensures that the number of bytes reported by GCS matches
+            the number we wrote; 'md5' does a full checksum. Any value other
+            than 'size' or 'md5' is assumed to mean no checking.
+        metadata: dict
+            Custom metadata, in key/value pairs, added at file creation
+        """
         bucket, key = split_path(path)
+        if not key:
+            raise OSError('Attempt to open a bucket')
         self.gcsfs = gcsfs
         self.bucket = bucket
         self.key = key
+        self.metadata = metadata
         self.mode = mode
         self.blocksize = block_size
         self.cache = b""
@@ -617,6 +765,9 @@ class GCSFile:
         self.start = None
         self.closed = False
         self.trim = True
+        self.consistency = consistency
+        if self.consistency == 'md5':
+            self.md5 = md5()
         if mode not in {'rb', 'wb'}:
             raise NotImplementedError('File mode not supported')
         if mode == 'rb':
@@ -629,6 +780,7 @@ class GCSFile:
             self.buffer = io.BytesIO()
             self.offset = 0
             self.forced = False
+            self.location = None
 
     def info(self):
         """ File information about this path """
@@ -716,6 +868,8 @@ class GCSFile:
             raise ValueError('File not in write mode')
         if self.closed:
             raise ValueError('I/O operation on closed file.')
+        if self.forced:
+            raise ValueError('This file has been force-flushed, can only close')
         out = self.buffer.write(ensure_writable(data))
         self.loc += out
         if self.buffer.tell() >= self.blocksize:
@@ -733,7 +887,7 @@ class GCSFile:
         ----------
         force : bool
             When closing, write the last block even if it is smaller than
-            blocks are allowed to be.
+            blocks are allowed to be. Disallows further writing to this file.
         """
         if self.mode not in {'wb', 'ab'}:
             raise ValueError('Flush on a file not in write mode')
@@ -744,21 +898,20 @@ class GCSFile:
             return
         if force and self.forced:
             raise ValueError("Force flush cannot be called more than once")
-        if force:
-            self.forced = True
         if force and not self.offset and self.buffer.tell() <= 5 * 2**20:
             self._simple_upload()
-            return
-
-        if not self.offset:
+        elif not self.offset:
             self._initiate_upload()
-        self._upload_chunk(final=force)
+        if self.location is not None:
+            self._upload_chunk(final=force)
+        if force:
+            self.forced = True
 
     def _upload_chunk(self, final=False):
         self.buffer.seek(0)
         data = self.buffer.read()
-        head = self.gcsfs.header.copy()
-        l = len(data)
+        head = {}
+        l = self.buffer.tell()
         if final:
             if l:
                 head['Content-Range'] = 'bytes %i-%i/%i' % (
@@ -772,42 +925,58 @@ class GCSFile:
                 self.offset, self.offset + l - 1)
         head.update({'Content-Type': 'application/octet-stream',
                      'Content-Length': str(l)})
-        r = requests.post(self.location, params={'uploadType': 'resumable'},
-                          headers=head, data=data)
+        r = self.gcsfs.session.post(
+            self.location, params={'uploadType': 'resumable'},
+            headers=head, data=data)
         validate_response(r, self.location)
         if 'Range' in r.headers:
+            assert not final, "Response looks like upload is partial"
             shortfall = (self.offset + l - 1) - int(
                     r.headers['Range'].split('-')[1])
             if shortfall:
+                if self.consistency == 'md5':
+                    self.md5.update(data[:-shortfall])
                 self.buffer = io.BytesIO(data[-shortfall:])
                 self.buffer.seek(shortfall)
             else:
+                if self.consistency == 'md5':
+                    self.md5.update(data)
                 self.buffer = io.BytesIO()
-            import pdb
-            # pdb.set_trace()
             self.offset += l - shortfall
         else:
+            assert final, "Response looks like upload is over"
+            size, md5 = int(r.json()['size']), r.json()['md5Hash']
+            if self.consistency == 'size':
+                assert size == self.buffer.tell() + self.offset, "Size mismatch"
+            if self.consistency == 'md5':
+                assert b64encode(
+                    self.md5.digest()) == md5.encode(), "MD5 checksum failed"
             self.buffer = io.BytesIO()
             self.offset += l
 
     def _initiate_upload(self):
-        r = requests.post('https://www.googleapis.com/upload/storage/v1/b/%s/o'
-                          % quote_plus(self.bucket),
-                          params={'uploadType': 'resumable'},
-                          headers=self.gcsfs.header, json={'name': self.key})
+        r = self.gcsfs.session.post(
+            'https://www.googleapis.com/upload/storage/v1/b/%s/o'
+            % quote_plus(self.bucket),
+            params={'uploadType': 'resumable'},
+            json={'name': self.key, 'metadata': self.metadata})
         self.location = r.headers['Location']
 
     def _simple_upload(self):
         """One-shot upload, less than 5MB"""
-        head = self.gcsfs.header.copy()
         self.buffer.seek(0)
         data = self.buffer.read()
         path = ('https://www.googleapis.com/upload/storage/v1/b/%s/o'
                 % quote_plus(self.bucket))
-        r = requests.post(path,
-                          params={'uploadType': 'media', 'name': self.key},
-                          headers=head, data=data)
+        r = self.gcsfs.session.post(
+            path, params={'uploadType': 'media', 'name': self.key}, data=data)
         validate_response(r, path)
+        size, md5 = int(r.json()['size']), r.json()['md5Hash']
+        if self.consistency == 'size':
+            assert size == self.buffer.tell(), "Size mismatch"
+        if self.consistency == 'md5':
+            self.md5.update(data)
+            assert b64encode(self.md5.digest()) == md5.encode(), "MD5 checksum failed"
 
     def _fetch(self, start, end):
         # force read to 5MB boundaries
@@ -817,20 +986,32 @@ class GCSFile:
             # First read
             self.start = start
             self.end = end + self.blocksize
-            self.cache = _fetch_range(self.gcsfs.header, self.details,
-                                      start, self.end)
+            self.cache = _fetch_range(self.details, self.gcsfs.session, start,
+                                      self.end)
         if start < self.start:
-            new = _fetch_range(self.gcsfs.header, self.details,
-                               start, self.start)
-            self.start = start
-            self.cache = new + self.cache
+            if self.end - end > self.blocksize:
+                self.start = start
+                self.end = end + self.blocksize
+                self.cache = _fetch_range(self.details, self.gcsfs.session,
+                                          self.start, self.end)
+            else:
+                new = _fetch_range(self.details, self.gcsfs.session, start,
+                                   self.start)
+                self.start = start
+                self.cache = new + self.cache
         if end > self.end:
             if self.end > self.size:
                 return
-            new = _fetch_range(self.gcsfs.header, self.details,
-                               self.end, end + self.blocksize)
-            self.end = end + self.blocksize
-            self.cache = self.cache + new
+            if end - self.end > self.blocksize:
+                self.start = start
+                self.end = end + self.blocksize
+                self.cache = _fetch_range(self.details, self.gcsfs.session,
+                                          self.start, self.end)
+            else:
+                new = _fetch_range(self.details, self.gcsfs.session, self.end,
+                                   end + self.blocksize)
+                self.end = end + self.blocksize
+                self.cache = self.cache + new
 
     def read(self, length=-1):
         """
@@ -896,45 +1077,46 @@ class GCSFile:
         self.close()
 
 
-def _fetch_range(head, obj_dict, start=None, end=None):
+def _fetch_range(obj_dict, session, start=None, end=None):
     """ Get data from GCS
 
-    head : dict
-        Contains authorization header
     obj_dict : an entry from ls() or info()
+    session: requests.Session instance
     start, end : None or integers
         if not both None, fetch only given range
     """
     if DEBUG:
         print('Fetch: ', start, end)
-    logger.debug("Fetch: {}/{}, {}-{}", obj_dict['name'], start, end)
+    logger.debug("Fetch: {}, {}-{}", obj_dict['name'], start, end)
     if start is not None or end is not None:
         start = start or 0
         end = end or 0
-        head = head.copy()
-        head['Range'] = 'bytes=%i-%i' % (start, end - 1)
-    back = requests.get(obj_dict['mediaLink'], headers=head)
+        head = {'Range': 'bytes=%i-%i' % (start, end - 1)}
+    else:
+        head = None
+    back = session.get(obj_dict['mediaLink'], headers=head)
     data = back.content
     if data == b'Request range not satisfiable':
         return b''
     return data
 
 
-def put_object(credentials, bucket, name, data):
+def put_object(credentials, bucket, name, data, session):
     """ Simple put, up to 5MB of data
 
     credentials : from auth()
     bucket : string
     name : object name
     data : binary
+    session: requests.Session instance
     """
-    out = requests.post('https://www.googleapis.com/upload/storage/'
-                        'v1/b/%s/o?uploadType=media&name=%s' % (
-                            quote_plus(bucket), quote_plus(name)),
-                        headers={'Authorization': 'Bearer ' +
-                                                  credentials.access_token,
-                                 'Content-Type': 'application/octet-stream',
-                                 'Content-Length': len(data)}, data=data)
+    out = session.post('https://www.googleapis.com/upload/storage/'
+                       'v1/b/%s/o?uploadType=media&name=%s' % (
+                           quote_plus(bucket), quote_plus(name)),
+                       headers={'Authorization': 'Bearer ' +
+                                                 credentials.access_token,
+                                'Content-Type': 'application/octet-stream',
+                                'Content-Length': len(data)}, data=data)
     assert out.status_code == 200
 
 
